""" A simulation Environment """

from pydantic import Extra

from .base_model import BaseModel


class Environment(BaseModel):
    """An environment contains all variables that don't depend on the
        geometry/structure itself such as most commonly wavelength and temperature.

    Attributes:
        wl: the wavelength of the environment
        T: the temperature of the environment
    """

    wl: float = 1.5
    T: float = 25.0

    class Config:
        allow_population_by_field_name = True
        extra = Extra.allow



""" Provides a custom pydantic BaseModel which handles numpy arrays better """

from typing import Any, Tuple

import numpy as np
from pydantic import validator
from pydantic.main import BaseModel as _BaseModel
from pydantic.main import ModelMetaclass as _ModelMetaclass


class _array(np.ndarray):
    """just an array with a nicer repr"""

    def __repr__(self):
        if self.ndim == 0:
            return f"{self:.3e}"
        cls_str = f"array{{{'x'.join(str(i) for i in self.shape)}}}"
        self = self.ravel()
        num_els = self.shape[0]
        if num_els == 0:
            return f"{cls_str}([])"
        elif num_els == 1:
            return f"{cls_str}([{self[0]:.3e}])"
        elif num_els == 2:
            return f"{cls_str}([{self[0]:.3e}, {self[1]:.3e}])"
        elif num_els == 3:
            return f"{cls_str}([{self[0]:.3e}, {self[1]:.3e}, {self[2]:.3e}])"
        else:
            return f"{cls_str}([{self[0]:.3e}, {self[1]:.3e}, ..., {self[-1]:.3e}])"


def _serialize_array(arr):
    if np.iscomplexobj(arr):
        return {
            "real": _serialize_array(np.real(arr)),
            "imag": _serialize_array(np.imag(arr)),
        }
    else:
        return np.round(arr, 12).tolist()


class ModelMetaclass(_ModelMetaclass):
    """A metaclass (used in our custom BaseModel) to handle numpy array type hints better.

    Generic numpy type hints (will) have the following syntax::

        arr: np.ndarray[Shape, DType]

    for example::

        arr: np.ndarray[Tuple[Literal[1], int], np.dtype[np.float64]] = np.array([[1, 2, 3]])

    When using this ModelMetaclass, the BaseModel will automatically try to cast the arrays according to the given array type-hint.

    Moreover, the following config will be injected (added to the config class that might already be present)::

        Config:
            arbitrary_types_allowed = True
            json_encoders = {
                np.ndarray: lambda arr: np.round(arr, 12).tolist()
            }

    """

    def __new__(cls, name, bases, dct, **kwargs):
        # Inject sensible default Configuration (Config class)
        extra_config = {
            "arbitrary_types_allowed": True,
            "json_encoders": {
                np.ndarray: _serialize_array,
                _array: _serialize_array,
                complex: lambda value: {"real": np.real(value), "imag": np.imag(value)},
            },
        }
        if "Config" not in dct:
            dct["Config"] = type("Config", (), extra_config)
        else:
            config = {**dct["Config"].__dict__}
            config["arbitrary_types_allowed"] = True
            config["json_encoders"] = {
                **config.get("json_encoders", {}),
                **extra_config["json_encoders"],
            }
            dct["Config"] = type("Config", (dct["Config"],), config)

        # Enforce numpy array annotations
        for attr, annot in dct.get("__annotations__", {}).items():
            if cls._is_array_annot(annot):

                @validator(attr, pre=True, allow_reuse=True, always=True)
                def validate(cls, x):

                    if isinstance(x, dict):
                        r = np.asarray(x.get("real", 0.0), dtype=np.float_)
                        i = np.asarray(x.get("imag", 0.0), dtype=np.float_)
                        x = r + i

                    shape, dtype = cls._parse_array_type_info(annot)
                    x = np.asarray(x, dtype=(None if dtype is Any else dtype))
                    if shape is not Any:
                        try:
                            shape = np.broadcast_shapes(shape, x.shape)
                        except ValueError:
                            raise ValueError(
                                f"Invalid shape for attribute 'x': Expected: {shape}. Got: {x.shape}."
                            )
                        x = np.broadcast_to(x, shape)
                    return x.view(_array)  # enables a more concise repr

                dct[f"validate_{attr}"] = validate

            if annot is complex:

                @validator(attr, pre=True, allow_reuse=True, always=True)
                def validate(cls, x):

                    if isinstance(x, dict):
                        r = np.asarray(x.get("real", 0.0), dtype=np.float_)
                        i = np.asarray(x.get("imag", 0.0), dtype=np.float_)
                        x = r + i

                    return float(np.real(x)) + 1j * float(np.imag(x))

                dct[f"validate_{attr}"] = validate

        return super().__new__(cls, name, bases, dct, **kwargs)

    @classmethod
    def _parse_array_type_info(cls, annotation) -> Tuple[Any, Any]:
        try:
            type_info = annotation.__args__
        except AttributeError:
            type_info = tuple()
        if len(type_info) == 1:
            (shape,) = type_info
            dtype = Any
        elif len(type_info) == 2:
            shape, dtype = type_info
        else:
            shape, dtype = Any, Any
        try:
            (dtype,) = dtype.__args__  # type: ignore
        except Exception:
            dtype = Any

        try:
            shape = shape.__args__  # type: ignore
            shape = tuple(cls._try_parse_shape_int(i) for i in shape)
        except Exception:
            shape = Any

        return shape, dtype

    @staticmethod
    def _try_parse_shape_int(value):
        from typing import _LiteralGenericAlias  # type: ignore

        if isinstance(value, _LiteralGenericAlias):
            (value,) = value.__args__
        try:
            return int(value)
        except Exception:
            return 1

    @staticmethod
    def _is_array_annot(annot):
        try:
            arr = annot(0)
        except Exception:
            return False
        return isinstance(arr, np.ndarray)


class BaseModel(_BaseModel, metaclass=ModelMetaclass):
    """A customized pydantic base model that handles numpy array type hints"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__dict__.update(_view_arrays(self.__dict__))

    def _repr(self, indent=0, shift=2):
        s = f"{self.__class__.__name__}("
        dic = self.dict()
        if dic:
            s = s + "\n"
        for key in dic:
            if key == "data":
                continue
            attr = getattr(self, key)
            if isinstance(attr, BaseModel):
                attr_str = attr._repr(indent=indent + shift, shift=shift)
            else:
                attr_str = repr(attr)
            s += f"{' '*(indent + shift)}{key}={attr_str}\n"
        s += f"{' '*indent})"
        return s

    def _visualize(self):
        raise NotImplementedError(
            f"visualization for {self.__class__.__name__!r} not (yet) implemented."
        )

    def __repr__(self):
        return self._repr()

    def __str__(self):
        return self._repr()


def _view_arrays(obj):
    if isinstance(obj, dict):
        return {k: _view_arrays(v) for k, v in obj.items()}
    elif isinstance(obj, np.ndarray):
        return obj.view(_array)
    else:
        return obj



""" compute modes with FDE """

from functools import wraps

from .lumerical import compute_modes as compute_modes_lumerical
from .tidy3d import compute_modes as compute_modes_tidy3d


@wraps(compute_modes_tidy3d)
def compute_modes(*args, **kwargs):
    return compute_modes_tidy3d(*args, **kwargs)



""" SAX backend for EME (default backend) """
from functools import partial
from typing import Any, Dict, List

import numpy as np
import sax

from ..mode import Mode, inner_product

try:
    import klujax
except ImportError:
    klujax = None


def compute_interface_s_matrix(
    modes1: List[Mode],
    modes2: List[Mode],
    enforce_reciprocity: bool = True,
    enforce_lossy_unitarity: bool = False,
):
    """get the S-matrix of the interface between two `CrossSections`"""
    # overlap matrices
    NL, NR = len(modes1), len(modes2)
    O_LL = np.array([inner_product(modes1[m], modes1[m]) for m in range(NL)])
    O_RR = np.array([inner_product(modes2[n], modes2[n]) for n in range(NR)])
    O_LR = np.array(
        [[inner_product(modes1[m], modes2[n]) for n in range(NR)] for m in range(NL)]
    )
    O_RL = np.array(
        [[inner_product(modes2[m], modes1[n]) for n in range(NL)] for m in range(NR)]
    )

    # extra phase correction.

    # ignoring the phase seems to corresponds best with lumerical.
    O_LL = np.abs(O_LL)
    O_RR = np.abs(O_RR)

    # alternative phase correction (probably worth testing this out)
    # O_LL = O_LL*np.exp(-1j*np.angle(O_LL))
    # O_RR = O_RR*np.exp(-1j*np.angle(O_RR))

    # yet another alternative phase correction (probably worth testing this out too)
    # O_LR = O_LR@np.diag(np.exp(-1j*np.angle(np.diag(O_LR))))
    # O_RL = O_RL@np.diag(np.exp(-1j*np.angle(np.diag(O_RL))))

    # transmission L->R
    LHS = O_LR + O_RL.T
    RHS = np.diag(2 * O_LL)
    T_LR = np.linalg.solve(LHS, RHS)
    U, t, V = np.linalg.svd(T_LR)

    # HACK: we don't expect gain --> invert singular values that lead to gain
    # see: https://github.com/BYUCamachoLab/emepy/issues/12
    t = np.where(t > 1, 1 / t, t)

    T_LR = U @ np.diag(t) @ V

    # transmission R->L
    LHS = O_RL + O_LR.T
    RHS = np.diag(2 * O_RR)
    T_RL = np.linalg.solve(LHS, RHS)
    U, t, V = np.linalg.svd(T_RL)

    # HACK: we don't expect gain --> invert singular values that lead to gain
    t = np.where(t > 1, 1 / t, t)

    T_RL = U @ np.diag(t) @ V

    # reflection
    R_LR = np.diag(1 / (2 * O_LL)) @ (O_RL.T - O_LR) @ T_LR  # type: ignore
    R_RL = np.diag(1 / (2 * O_RR)) @ (O_LR.T - O_RL) @ T_RL  # type: ignore

    # s-matrix
    S = np.concatenate(
        [
            np.concatenate([R_LR, T_RL], 1),
            np.concatenate([T_LR, R_RL], 1),
        ],
        0,
    )

    # enforce S@S.H is diagonal
    if enforce_lossy_unitarity:  # HACK!
        U, s, V = np.linalg.svd(S)
        S = np.diag(s) @ U @ V

    # ensure reciprocity:
    if enforce_reciprocity:
        S = 0.5 * (S + S.T)

    # create port map
    in_ports = [f"left@{i}" for i in range(len(modes1))]
    out_ports = [f"right@{i}" for i in range(len(modes2))]
    port_map = {p: i for i, p in enumerate(in_ports + out_ports)}

    return S, port_map


def compute_interface_s_matrices(
    modes: List[List[Mode]],
    enforce_reciprocity: bool = True,
    enforce_lossy_unitarity: bool = False,
):
    """get all the S-matrices of all the interfaces in a collection of `CrossSections`"""
    return {
        f"i_{i}_{i + 1}": compute_interface_s_matrix(
            modes1=modes1,
            modes2=modes2,
            enforce_reciprocity=enforce_reciprocity,
            enforce_lossy_unitarity=enforce_lossy_unitarity,
        )
        for i, (modes1, modes2) in enumerate(zip(modes[:-1], modes[1:]))
    }


def compute_propagation_s_matrix(modes: List[Mode]):
    """get the propagation S-matrix of each `Mode` belonging to a `CrossSection` in a `Cell` with a certain length."""
    s_dict = {
        (f"left@{i}", f"right@{i}"): np.exp(
            2j * np.pi * np.abs(mode.neff) / mode.env.wl * mode.cell.length
        )
        for i, mode in enumerate(modes)
    }
    s_dict = {**s_dict, **{(p2, p1): v for (p1, p2), v in s_dict.items()}}
    return s_dict


def compute_propagation_s_matrices(modes: List[List[Mode]]):
    """get all the propagation S-matrices of all the `Modes` belonging to each `CrossSection`"""
    return {
        f"p_{i}": compute_propagation_s_matrix(modes_)
        for i, modes_ in enumerate(modes)
    }


def get_netlist(propagations, interfaces):
    """get the netlist of a stack of `Modes`"""

    instances = {
        **{k: _load_constant_model(S) for k, S in propagations.items()},
        **{k: _load_constant_model(S) for k, S in interfaces.items()},
    }

    connections = {}
    for i in range(len(interfaces)):
        connections[f"p_{i},right"] = f"i_{i}_{i+1},left"
        connections[f"i_{i}_{i+1},right"] = f"p_{i+1},left"

    ports = {
        f"left": f"p_0,left",
        f"right": f"p_{len(propagations)-1},right",
    }

    return {"instances": instances, "connections": connections, "ports": ports}


def _load_constant_model(value):
    def model():
        return value

    return model


def _validate_sax_backend(sax_backend):
    if sax_backend is None:
        sax_backend = "klu" if klujax is not None else "default"

    if sax_backend not in ["default", "klu"]:
        raise ValueError(
            f"Invalid SAX Backend. Got: {sax_backend!r}. Should be 'default' or 'klu'."
        )
    return sax_backend


def select_ports(
    S: np.ndarray[Any, np.dtype[np.float_]], port_map: Dict[str, int], ports: List[str]
):
    """Keep subset of an S-matrix

    Args:
        S: the S-matrix to downselect from
        port_map: a port name to s-matrix index mapping
        ports: the port names to keep

    Returns:
        the downselected s-matrix and port map
    """
    idxs = np.array([port_map[port] for port in ports], dtype=np.int_)
    s = S[idxs, :][:, idxs]
    new_port_map = {p: i for i, p in enumerate(ports)}
    return s, new_port_map


def compute_s_matrix(
    modes: List[List[Mode]],
    sax_backend=None,
    enforce_reciprocity=True,
    enforce_lossy_unitarity=False,
    **kwargs,
):
    """Calculate the S-matrix for given sets of modes, each set belonging to a `Cell`

    Args:
        modes: Each collection of modes for each of the `Cells`
        backend: which SAX backend to use to calculate the final S-matrix.
    """
    sax_backend = _validate_sax_backend(sax_backend)
    get_interface_s_matrices_ = kwargs.pop(
        "get_interface_s_matrices", compute_interface_s_matrices
    )
    get_interface_s_matrices = partial(get_interface_s_matrices_, **kwargs)

    num_modes = len(modes[0])
    propagations = compute_propagation_s_matrices(modes)
    interfaces = get_interface_s_matrices(
        modes,
        enforce_reciprocity=enforce_reciprocity,
        enforce_lossy_unitarity=enforce_lossy_unitarity,
    )
    net = get_netlist(propagations, interfaces)
    mode_names = tuple(f"{i}" for i in range(num_modes))

    _circuit = sax.circuit(**net, backend=sax_backend, modes=mode_names)

    def eme_model():
        sdict = sax.sdict(_circuit())
        sdict = {k: sdict[k] for k in sorted(sdict)}
        return sax.sdense(sdict)

    # maybe in the future we should return the sax model and not the S-matrix?
    return eme_model()



""" Visualizations for common meow-datatypes """

from typing import Any

import matplotlib.pyplot as plt
import numpy as np

try:
    import gdsfactory as gf
except ImportError:
    gf = None

try:
    from jaxlib.xla_extension import DeviceArray
except ImportError:
    DeviceArray = None


def _visualize_s_matrix(S, fmt=".3f", title=None):
    Z = np.abs(S)
    y, x = np.arange(Z.shape[0])[::-1], np.arange(Z.shape[1])
    Y, X = np.meshgrid(y, x)

    plt.figure(figsize=(2 * x.shape[0] / 3, 2 * y.shape[0] / 3))
    plt.pcolormesh(X, Y, Z[::-1].T, cmap="Greys", vmin=0.0, vmax=2.0 * Z.max())

    coords_ = np.concatenate(
        [np.array([x[0] - 1]), x, np.array([x[-1] + 1])], axis=0, dtype=float
    )
    labels = ["" for _ in coords_]
    plt.xticks(coords_ + 0.5, labels)
    plt.xlim(coords_[0] + 0.5, coords_[-1] - 0.5)

    coords_ = np.concatenate(
        [np.array([y[0] + 1]), y, np.array([y[-1] - 1])], axis=0, dtype=float
    )
    coords_ = coords_[::-1]  # reverse
    labels = ["" for _ in coords_]
    plt.yticks(coords_ + 0.5, labels)
    plt.ylim(coords_[-1] - 0.5, coords_[0] + 0.5)
    plt.grid(True)

    for x, y, z in zip(X.ravel(), Y.ravel(), Z[::-1].T.ravel()):
        if z > 0.0005:
            text = eval(f"f'{{z:{fmt}}}'")  # 😅
            plt.text(x, y, text, ha="center", va="center")

    if title is not None:
        plt.title(title)

    plt.show()


def _visualize_gdsfactory(comp):
    import gdsfactory as gf

    gf.plot(comp)  # type: ignore


def visualize(obj: Any, **kwargs: Any):
    """visualize any meow object

    Args:
        obj: the meow object to visualize
        **kwargs: extra configuration to visualize the object

    Note:
        Most meow objects have a `._visualize` method.
        Check out its help to see which kwargs are accepted.
    """
    from .base_model import BaseModel
    from .structures import Structure, _visualize_structures

    if isinstance(obj, BaseModel):
        return obj._visualize(**kwargs)
    elif (
        (
            isinstance(obj, np.ndarray)
            or (DeviceArray is not None and isinstance(obj, DeviceArray))
        )
        and obj.ndim == 2
        and obj.shape[0] > 1
        and obj.shape[1] > 1
    ):
        _visualize_s_matrix(obj, **kwargs)
    elif gf is not None and isinstance(obj, gf.Component):
        _visualize_gdsfactory(obj, **kwargs)
    else:
        try:
            (*objs,) = obj  # type: ignore
        except TypeError:
            return obj

        if all(isinstance(obj, Structure) for obj in objs):
            return _visualize_structures(objs, **kwargs)
        return objs


vis = visualize  # shorthand for visualize
